use proc_macro::TokenStream;
use proc_macro2::{Ident, TokenStream as TokenStream2};
use quote::quote;
use std::{collections::HashMap, str::FromStr};
use strum::EnumString;
use syn::{
    parse_macro_input, punctuated::Punctuated, AttributeArgs, FnArg, ItemFn, Lit, NestedMeta,
    ReturnType, Signature, Token, Type,
};

#[derive(EnumString, Debug)]
enum ReqMethod {
    GET,
    POST,
}

#[derive(Debug)]
struct Endpoint {
    method: ReqMethod,
}

#[derive(Debug)]
struct FnArgs {
    url_param: Option<Type>,
    payload: Option<Type>,
}

#[proc_macro_attribute]
pub fn api_route(args: TokenStream, item: TokenStream) -> TokenStream {
    let args: AttributeArgs = parse_macro_input!(args);
    let input: ItemFn = parse_macro_input!(item);
    let Signature {
        ident,
        inputs,
        generics,
        output,
        ..
    } = input.sig;
    let (impl_generics, _, _) = generics.split_for_impl();
    let FnArgs { url_param, payload } = parse_fn_args(inputs);
    let Endpoint { method } = parse_attr_args(args);
    let url_path = build_url_path(&ident, &url_param);
    let return_ty = match output {
        ReturnType::Type(_, ty) => quote!(#ty),
        ReturnType::Default => quote!(()),
    };
    let request = match method {
        ReqMethod::GET => build_get(url_path),
        ReqMethod::POST => build_post(url_path, &payload),
    };
    let url_param = match url_param {
        Some(p) => quote!(, param: #p),
        None => TokenStream2::new(),
    };
    let payload = match payload {
        Some(p) => quote!(, payload: #p),
        None => TokenStream2::new(),
    };
    let q = quote! {
        pub async fn #ident #impl_generics(
            &self #url_param #payload
        ) -> UResult<#return_ty> {
            let request = {
                #request
            };
            let response = request.send().await?;
            let content_len = response.content_length();
            let is_success = match response.error_for_status_ref() {
                Ok(_) => Ok(()),
                Err(e) => Err(UError::from(e))
            };
            match is_success {
                Ok(_) => response.json::<BaseMessage<#return_ty>>()
                    .await
                    .map(|msg| msg.into_inner())
                    .or_else(|e| {
                        match content_len {
                            Some(0) => Ok(Default::default()),
                            _ => Err(UError::from(e))
                        }
                    }),
                Err(UError::NetError(err_src, _)) => Err(
                    UError::NetError(
                        err_src,
                        response.text().await.unwrap()
                    )
                ),
                _ => unreachable!()
            }
        }
    };
    //eprintln!("#!#! RESULT:\n{}", q);
    q.into()
}

fn parse_fn_args(raw: Punctuated<FnArg, Token![,]>) -> FnArgs {
    let mut arg: HashMap<String, Type> = raw
        .into_iter()
        .filter_map(|arg| {
            if let FnArg::Typed(argt) = arg {
                let mut arg_name = String::new();
                // did you think I won't overplay you? won't destroy?
                |arg_ident| -> TokenStream {
                    let q: TokenStream = quote!(#arg_ident).into();
                    arg_name = parse_macro_input!(q as Ident).to_string();
                    TokenStream::new()
                }(argt.pat);
                if &arg_name != "url_param" && &arg_name != "payload" {
                    panic!("Wrong arg name: {}", &arg_name)
                }
                let arg_type = *argt.ty.clone();
                Some((arg_name, arg_type))
            } else {
                None
            }
        })
        .collect();
    FnArgs {
        url_param: arg.remove("url_param"),
        payload: arg.remove("payload"),
    }
}

fn build_get(url: TokenStream2) -> TokenStream2 {
    quote! {
        let request = self.build_get(#url);
        request
    }
}

fn build_post(url: TokenStream2, payload: &Option<Type>) -> TokenStream2 {
    let pld = match payload {
        Some(_) => quote! {
            .json(&payload.as_message())
        },
        None => TokenStream2::new(),
    };
    quote! {
        let request = self.build_post(#url);
        request #pld
    }
}

fn build_url_path(path: &Ident, url_param: &Option<Type>) -> TokenStream2 {
    let url_param = match url_param {
        Some(_) => quote! {
            + &opt_to_string(param)
        },
        None => TokenStream2::new(),
    };
    quote! {
        &format!(
            "{}/{}",
            stringify!(#path),
            String::new() #url_param
        )
    }
}

fn parse_attr_args(args: AttributeArgs) -> Endpoint {
    let mut args = args.into_iter();
    let method = match args.next() {
        Some(method) => match method {
            NestedMeta::Lit(l) => {
                if let Lit::Str(s) = l {
                    match ReqMethod::from_str(&s.value()) {
                        Ok(v) => v,
                        Err(_) => panic!("Unknown method"),
                    }
                } else {
                    panic!("Method must be a str")
                }
            }
            _ => panic!("Method must be on the first place"),
        },
        None => panic!("Method required"),
    };
    Endpoint { method }
}
