use super::JobState;
use crate::{
    cache::JobCache,
    models::{schema::*, ExecResult, JobOutput},
    utils::{systime_to_string, TempFile},
    UID,
};
use diesel::{Identifiable, Insertable, Queryable};
use serde::{Deserialize, Serialize};
use std::{fmt, process::Output, string::FromUtf8Error, time::SystemTime};
use tokio::process::Command;
use uuid::Uuid;

#[derive(
    Serialize, Deserialize, Clone, Debug, Queryable, Identifiable, Insertable, AsChangeset,
)]
#[table_name = "results"]
pub struct AssignedJob {
    pub agent_id: Uuid,
    pub alias: Option<String>,
    pub created: SystemTime,
    pub id: Uuid,
    pub job_id: Uuid,
    pub result: Option<Vec<u8>>,
    pub state: JobState,
    pub retcode: Option<i32>,
    pub updated: SystemTime,
}

impl fmt::Display for AssignedJob {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        let mut out = format!("Result: {}", self.id);
        out += &format!("\nAgent: {}", self.agent_id);
        out += &format!("\nJob: {}", self.job_id);
        if self.alias.is_some() {
            out += &format!("\nAlias: {}", self.alias.as_ref().unwrap());
        }
        out += &format!("\nUpdated: {}", systime_to_string(&self.updated));
        out += &format!("\nState: {}", self.state);
        if self.state == JobState::Finished {
            if self.retcode.is_some() {
                out += &format!("\nReturn code: {}", self.retcode.unwrap());
            }
            if self.result.is_some() {
                out += &format!(
                    "\nResult: {}",
                    String::from_utf8_lossy(self.result.as_ref().unwrap())
                );
            }
        }
        write!(f, "{}", out)
    }
}

impl Default for AssignedJob {
    fn default() -> Self {
        Self {
            agent_id: Uuid::nil(),
            alias: None,
            created: SystemTime::now(),
            id: Uuid::new_v4(),
            job_id: Uuid::nil(),
            result: None,
            state: JobState::Queued,
            retcode: None,
            updated: SystemTime::now(),
        }
    }
}

impl AssignedJob {
    pub async fn run(mut self) -> ExecResult {
        let (argv, _payload) = {
            let meta = JobCache::get(&self.job_id).unwrap();
            let extracted_payload = meta
                .payload
                .as_ref()
                .and_then(|p| TempFile::write_exec(p).ok());
            let argv = if let Some(ref p) = &extracted_payload {
                meta.argv.replace("{}", &p.get_path())
            } else {
                meta.argv.clone()
            };
            (argv, extracted_payload)
        };
        let mut split_cmd = shlex::split(&argv).unwrap().into_iter();
        let cmd = split_cmd.nth(0).unwrap();
        let args = split_cmd.collect::<Vec<String>>();
        let cmd_result = Command::new(cmd).args(args).output().await;
        let (data, retcode) = match cmd_result {
            Ok(Output {
                status,
                stdout,
                stderr,
            }) => (
                JobOutput::new()
                    .stdout(stdout)
                    .stderr(stderr)
                    .into_combined(),
                status.code(),
            ),
            Err(e) => (
                JobOutput::new()
                    .stderr(e.to_string().into_bytes())
                    .into_combined(),
                None,
            ),
        };
        self.result = Some(data);
        self.retcode = retcode;
        self.updated = SystemTime::now();
        self.state = JobState::Finished;
        ExecResult::Assigned(self)
    }

    pub fn new(job_id: Uuid, other: Option<&Self>) -> Self {
        Self {
            agent_id: *UID,
            job_id,
            ..other.unwrap_or(&Default::default()).clone()
        }
    }

    pub fn as_job_output(&self) -> Option<JobOutput> {
        self.result
            .as_ref()
            .and_then(|r| JobOutput::from_combined(r))
    }

    pub fn to_raw_result(&self) -> Vec<u8> {
        match self.result.as_ref() {
            Some(r) => match JobOutput::from_combined(r) {
                Some(o) => o.to_appropriate(),
                None => r.clone(),
            },
            None => b"No data".to_vec(),
        }
    }

    pub fn to_string_result(&self) -> Result<String, FromUtf8Error> {
        String::from_utf8(self.to_raw_result())
    }
}
